package cds;

import java.util.concurrent.locks.*;

/**
 * CondBoundedBuffer
 * <p/>
 * Bounded buffer using \Condition
 *
 * This implementation has been obtained by reworking a related class by Brian
 * Goetz and Tim Peierls
 *
 * @param <T>
 */
public class CondBoundedBuffer<T> implements BoundedBufferInterface<T> {

    private final Lock mutex = new ReentrantLock();
    private final Condition notFull  = mutex.newCondition();
    private final Condition notEmpty = mutex.newCondition();

    private final T[] buffer;
    private int putIndex, takeIndex;
    private volatile int count;

    public CondBoundedBuffer(int capacity) {
        if (capacity <= 0) {
            throw new IllegalArgumentException();
        }
        buffer = (T[]) new Object[capacity]; // just a trick
    }

    @Override
    public void put(T x) throws InterruptedException {
        mutex.lock();
        try {
            while (count == buffer.length) {
                notFull.await();
            }
            buffer[putIndex] = x;
            if (++putIndex == buffer.length) {
                putIndex = 0;
            }
            //putIndex = ++putIndex%buffer.length;
            ++count;
            notEmpty.signal();
        } finally {
            mutex.unlock();
        }
    }

    @Override
    public T take() throws InterruptedException {
        mutex.lock();
        try {
            while (count == 0) {
                notEmpty.await();
            }
            T x = buffer[takeIndex];
            if (++takeIndex == buffer.length) {
                takeIndex = 0;
            }
            //takeIndex=++takeIndex%buffer.length;
            --count;
            notFull.signal();
            return x;
        } finally {
            mutex.unlock();
        }
    }

}
